from abc import ABCMeta, abstractmethod


class RpcParser(metaclass=ABCMeta):

    @abstractmethod
    def convert_to_rpc_request(*args, **kwargs) -> dict:
        raise NotImplementedError


    @abstractmethod
    def convert_from_rpc_response(rpc_response: dict) -> dict:
        raise NotImplementedError


    @abstractmethod
    def parse_and_throw_error(response: dict) -> None:
        raise NotImplementedError


    def is_error(response: dict) -> bool:
        return 'error' in response




class TonConnectError(Exception):
    prefix = '[TON_CONNECT_SDK_ERROR]'
    info = None

    def __init__(self, message=None):
        super(TonConnectError, self).__init__(f'{self.prefix}'
                                              + (f': {self.info}' if self.info else '')
                                              + (f' {message}' if message is not None else '')
                                              )


class WalletAlreadyConnectedError(TonConnectError):
    info = 'Wallet connection called but wallet already connected. To avoid the error, disconnect the wallet before doing a new connection.'


class WalletNotConnectedError(TonConnectError):
    info = 'Send transaction or other protocol methods called while wallet is not connected.'


class WalletNotSupportFeatureError(TonConnectError):
    info = "Wallet doesn't support requested feature method."


class FetchWalletsError(TonConnectError):
    info = 'An error occurred while fetching the wallets list.'


class UnknownError(TonConnectError):
    info = 'Unknown error.'


class BadRequestError(TonConnectError):
    info = 'Request to the wallet contains errors.'


class UnknownAppError(TonConnectError):
    info = 'App tries to send rpc request to the injected wallet while not connected.'


class UserRejectsError(TonConnectError):
    info = 'User rejects the action in the wallet.'


class ManifestNotFoundError(TonConnectError):
    info = 'Manifest not found. Make sure you added `tonconnect-manifest.json` to the root of your app or passed correct manifest_url. See more https://github.com/ton-connect/docs/blob/main/requests-responses.md#app-manifest'


class ManifestContentError(TonConnectError):
    info = 'Passed `tonconnect-manifest.json` contains errors. Check format of your manifest. See more https://github.com/ton-connect/docs/blob/main/requests-responses.md#app-manifest'



import asyncio
import json
from urllib.parse import quote_plus

from pytonconnect.crypto import SessionCrypto
from pytonconnect.exceptions import TonConnectError
from pytonconnect.logger import _LOGGER
from pytonconnect.storage import IStorage

from ._provider import BaseProvider
from ._bridge_gateway import BridgeGateway
from ._bridge_session import BridgeSession


class BridgeProvider(BaseProvider):

    DISCONNECT_TIMEOUT = 600
    STANDART_UNIVERSAL_URL = 'tc://'

    _storage: IStorage
    _wallet: dict

    _session: BridgeSession
    _gateway: BridgeGateway
    _pending_requests: dict
    _listeners: list


    def __init__(self, storage: IStorage, wallet: dict = None):
        self._storage = storage
        self._wallet = wallet

        self._session = BridgeSession()
        self._gateway = None
        self._pending_requests = {}
        self._listeners = []


    async def connect(self, request: dict):
        self._close_gateways()
        session_crypto = SessionCrypto()
        
        bridge_url = ''
        universal_url = BridgeProvider.STANDART_UNIVERSAL_URL

        if isinstance(self._wallet, dict):
            bridge_url = self._wallet['bridge_url']
            if 'universal_url' in self._wallet:
                universal_url = self._wallet['universal_url']
            
            self._gateway = BridgeGateway(
                self._storage,
                bridge_url,
                session_crypto.session_id,
                self._gateway_listener,
                self._gateway_errors_listener
            )

            await self._gateway.register_session()
        
        self._session.session_crypto = session_crypto
        self._session.bridge_url = bridge_url

        return self._generate_universal_url(universal_url, request)


    async def restore_connection(self):
        self._close_gateways()

        connection = await self._storage.get_item(IStorage.KEY_CONNECTION)
        if not connection:
            return False
        connection = json.loads(connection)

        if 'session' not in connection:
            return False
        self._session = BridgeSession(connection['session'])
        
        self._gateway = BridgeGateway(
            self._storage,
            self._session.bridge_url,
            self._session.session_crypto.session_id,
            self._gateway_listener,
            self._gateway_errors_listener
        )

        await self._gateway.register_session()

        for listener in self._listeners:
            listener(connection['connect_event'])
        
        return True


    def close_connection(self):
        self._close_gateways()
        self._session = BridgeSession()
        self._gateway = None
        self._pending_requests = {}
        self._listeners = []


    async def disconnect(self):
        loop = asyncio.get_running_loop()
        resolve = loop.create_future()

        def on_request_sent(request_future: asyncio.Future):
            asyncio.create_task(self._remove_session()) \
                    .add_done_callback(lambda x: resolve.set_result(True) if not resolve.done() else None)
            request_future.set_result(None)

        try:
            await asyncio.wait_for(self.send_request({'method': 'disconnect', 'params': []},
                                                     on_request_sent=on_request_sent),
                                                     timeout=self.DISCONNECT_TIMEOUT)
        except Exception:
            _LOGGER.exception(f'Provider disconnect')
        finally:
            if not resolve.done():
                await self._remove_session()
                resolve.set_result(True)

        return await resolve


    def pause(self):
        if self._gateway is not None:
            self._gateway.pause()


    async def unpause(self):
        if self._gateway is not None:
            await self._gateway.unpause()


    async def send_request(self, request: dict, on_request_sent = None):
        if not self._gateway or not self._session or not self._session.wallet_public_key:
            raise TonConnectError('Trying to send bridge request without session.')

        connection = json.loads(await self._storage.get_item(IStorage.KEY_CONNECTION, '{}'))
        id = connection.get('next_rpc_request_id', '0')
        connection['next_rpc_request_id'] = str(int(id)+1)
        await self._storage.set_item(IStorage.KEY_CONNECTION, json.dumps(connection))

        request['id'] = id
        _LOGGER.debug(f'Provider send http-bridge request: {request}')

        encoded_request = self._session.session_crypto.encrypt(
            json.dumps(request),
            self._session.wallet_public_key
        )

        await self._gateway.send(encoded_request, self._session.wallet_public_key, request['method'])

        loop = asyncio.get_running_loop()
        resolve = loop.create_future()

        self._pending_requests[id] = resolve
        if on_request_sent is not None:
            on_request_sent(resolve)

        return await resolve


    def listen(self, callback):
        self._listeners.append(callback)


    async def _gateway_listener(self, bridge_incoming_message):
        wallet_message = json.loads(
            self._session.session_crypto.decrypt(bridge_incoming_message['message'], bridge_incoming_message['from'])
        )

        _LOGGER.debug(f'Wallet message received: {wallet_message}')

        if 'event' not in wallet_message:
            if 'id' in wallet_message:
                id = wallet_message['id']
                if id not in self._pending_requests:
                    _LOGGER.debug(f"Response id {id} doesn't match any request's id")
                    return
                
                self._pending_requests[id].set_result(wallet_message)
                del self._pending_requests[id]
            return

        if 'id' in wallet_message:
            id = int(wallet_message['id'])
            connection = json.loads(await self._storage.get_item(IStorage.KEY_CONNECTION, '{}'))
            last_id = connection['last_wallet_event_id'] if 'last_wallet_event_id' in connection else 0

            if last_id and id <= last_id:
                _LOGGER.error(f'Received event id (={id}) must be greater than stored last wallet event id (={last_id})')
                return
            
            if 'event' in wallet_message and wallet_message['event'] != 'connect':
                connection['last_wallet_event_id'] = id
                await self._storage.set_item(IStorage.KEY_CONNECTION, json.dumps(connection))

        # self.listeners might be modified in the event handler
        listeners = self._listeners.copy()

        if wallet_message['event'] == 'connect':
            await self._update_session(wallet_message, bridge_incoming_message['from'])

        elif wallet_message['event'] == 'disconnect':
            await self._remove_session()
        
        for listener in listeners:
            listener(wallet_message)


    async def _gateway_errors_listener(self, e = None):
        raise TonConnectError(f'Bridge error {json.dumps(e or {})}')


    async def _update_session(self, connect_event: dict, wallet_public_key: str):
        self._session.wallet_public_key = wallet_public_key

        connection = {
            'type': 'http',
            'session': self._session.get_dict(),
            'last_wallet_event_id': connect_event['id'] if 'id' in connect_event else None,
            'connect_event': connect_event,
            'next_rpc_request_id': 0
        }

        await self._storage.set_item(IStorage.KEY_CONNECTION, json.dumps(connection))


    async def _remove_session(self):
        if self._gateway is not None:
            self.close_connection()
            await self._storage.remove_item(IStorage.KEY_CONNECTION)
            await self._storage.remove_item(IStorage.KEY_LAST_EVENT_ID)


    def _generate_universal_url(self, universal_url: str, request: dict):
        version = 2
        session_id = self._session.session_crypto.session_id
        request_safe = quote_plus(json.dumps(request))

        universal_base = universal_url.rstrip('/')
        url = f'{universal_base}?v={version}&id={session_id}&r={request_safe}'

        return url


    def _close_gateways(self):
        if self._gateway:
            self._gateway.close()



from abc import ABCMeta, abstractmethod


class IStorage(metaclass=ABCMeta):

    KEY_LAST_EVENT_ID = 'last_event_id'
    KEY_CONNECTION = 'connection'


    @abstractmethod
    async def set_item(self, key: str, value: str):
        """Saves the `value` to the storage. Value can be accessed later by the `key`.
        
        :param key: key to access to the value later
        :param value: value to save
        """
        raise NotImplementedError


    @abstractmethod
    async def get_item(self, key: str, default_value: str = None):
        """Reads the `value` from the storage.
        
        :param key: key to access the value
        :param default_value: default value if key not found in storage
        """
        raise NotImplementedError
    

    @abstractmethod
    async def remove_item(self, key: str):
        """Removes the `value` from the storage.

        :param key: key to remove the value
        """
        raise NotImplementedError



import requests
from datetime import datetime

from pytonconnect.exceptions import FetchWalletsError
from pytonconnect.logger import _LOGGER


FALLBACK_WALLETS_LIST = [
    {
        "name": "Tonkeeper",
        "image": "https://tonkeeper.com/assets/tonconnect-icon.png",
        "tondns": "tonkeeper.ton",
        "about_url": "https://tonkeeper.com",
        "universal_url": "https://app.tonkeeper.com/ton-connect",
        "bridge": [
            {
                "type": "sse",
                "url": "https://bridge.tonapi.io/bridge"
            },
            {
                "type": "js",
                "key": "tonkeeper"
            }
        ]
    },
    {
        "name": "Tonhub",
        "image": "https://tonhub.com/tonconnect_logo.png",
        "about_url": "https://tonhub.com",
        "universal_url": "https://tonhub.com/ton-connect",
        "bridge": [
            {
                "type": "js",
                "key": "tonhub"
            },
            {
                "type": "sse",
                "url": "https://connect.tonhubapi.com/tonconnect"
            }
        ]
    }
]


class WalletsListManager:

    _wallets_list_source = 'https://raw.githubusercontent.com/ton-blockchain/wallets-list/main/wallets.json'
    _cache_ttl: int

    _wallets_list_cache: dict
    _wallets_list_cache_creation_timestamp: int


    def __init__(self, wallets_list_source = None, cache_ttl = None):
        if wallets_list_source:
            self._wallets_list_source = wallets_list_source
        self._cache_ttl = cache_ttl

        self._wallets_list_cache = None
        self._wallets_list_cache_creation_timestamp = None


    def get_wallets(self):
        if self._cache_ttl \
                and self._wallets_list_cache_creation_timestamp \
                and int(datetime.now().timestamp()) > self._wallets_list_cache_creation_timestamp + self._cache_ttl:
            self._wallets_list_cache = None

        if not self._wallets_list_cache:
            wallets_list = None
            try:
                wallets_list = requests.get(self._wallets_list_source).json()
                if not isinstance(wallets_list, list):
                    raise FetchWalletsError('Wrong wallets list format, wallets list must be an array.')

            except Exception as e:
                _LOGGER.error(f'WalletsListManager get_wallets {type(e)}: {e}')
                wallets_list = FALLBACK_WALLETS_LIST

            self._wallets_list_cache = []
            for wallet in wallets_list:
                supported_wallet = self._get_supported_wallet_config(wallet)
                if supported_wallet:
                    self._wallets_list_cache.append(supported_wallet)

            self._wallets_list_cache_creation_timestamp = int(datetime.now().timestamp())

        return self._wallets_list_cache


    def _get_supported_wallet_config(self, wallet):
        if not isinstance(wallet, dict):
            _LOGGER.warning(f'Not supported wallet: is not a dict -> {wallet}')
            return None

        containsName = 'name' in wallet
        containsImage = 'image' in wallet
        containsAbout = 'about_url' in wallet

        if not containsName or not containsImage or not containsAbout:
            _LOGGER.warning(f'Not supported wallet: contains -> name({containsName}), image({containsImage}), about({containsAbout}), config -> {wallet}')
            return None

        if 'bridge' not in wallet or not isinstance(wallet['bridge'], list) or not len(wallet['bridge']):
            _LOGGER.warning(f'Not supported wallet: bridge is not a list or len is equal 0, config -> {wallet}')
            return None

        walletConfig = {
            'name': wallet['name'],
            'image': wallet['image'],
            'about_url': wallet['about_url']
        }

        for bridge in wallet['bridge']:
            if bridge['type'] == 'sse':
                if 'url' not in bridge:
                    _LOGGER.warning(f'Not supported wallet: bridge url not found, config -> {wallet}')
                    return None

                walletConfig['bridge_url'] = bridge['url']
                if 'universal_url' in wallet:
                    walletConfig['universal_url'] = wallet['universal_url']
                break
        
        if 'bridge_url' not in walletConfig:
            _LOGGER.warning(f'Not supported wallet: sse not found, config -> {wallet}')
            return None

        return walletConfig


if __name__ == '__main__':
    wallets_list = WalletsListManager()
    print('Supported list of wallets:', wallets_list.get_wallets())



import json
from pytonconnect.crypto import SessionCrypto


class BridgeSession:

    session_crypto: SessionCrypto
    wallet_public_key: str
    bridge_url: str


    def __init__(self, stored: dict = None):
        self.session_crypto = SessionCrypto(stored['session_private_key']) if stored and 'session_private_key' in stored else None
        self.wallet_public_key = stored['wallet_public_key'] if stored and 'wallet_public_key' in stored else None
        self.bridge_url = stored['bridge_url'] if stored and 'bridge_url' in stored else None


    def __repr__(self):
        return json.dumps(self.get_dict())


    def get_dict(self):
        return {
            'session_private_key': self.session_crypto.key_pair.encode().hex(),
            'wallet_public_key': self.wallet_public_key,
            'bridge_url': self.bridge_url
        }


from ._send_transaction import SendTransactionParser
from ._connect_event import ConnectEventParser, WalletInfo, DeviceInfo, Account, TonProof

__all__ = [
    'SendTransactionParser',

    'ConnectEventParser',
    'WalletInfo',
    'DeviceInfo',
    'Account',
    'TonProof',
]



from base64 import b64decode, b64encode
from nacl.utils import random
from nacl.public import PrivateKey, PublicKey, Box
from nacl.encoding import HexEncoder


class SessionCrypto:

    key_pair: PrivateKey
    session_id: str


    def __init__(self, private_key: str = None):
        self.key_pair = PrivateKey(private_key, HexEncoder) if private_key else PrivateKey.generate()
        self.session_id = self.key_pair.public_key.encode().hex()
    

    def create_nonce(self):
        return random(Box.NONCE_SIZE)


    def encrypt(self, message: str, receiver_pub_key_hex: str):
        nonce = self.create_nonce()

        receiver_pk = PublicKey(receiver_pub_key_hex, HexEncoder)
        box = Box(self.key_pair, receiver_pk)
        encrypted = box.encrypt(message.encode('utf-8'), nonce)

        res = bytearray(nonce)
        res.extend(encrypted.ciphertext)
        return b64encode(bytes(res))


    def decrypt(self, message: bytes, sender_pub_key_hex: str):
        msg = b64decode(message)
        nonce = msg[:Box.NONCE_SIZE]
        internal_message = msg[Box.NONCE_SIZE:]

        sender_pk = PublicKey(sender_pub_key_hex, HexEncoder)
        box = Box(self.key_pair, sender_pk)

        decrypted = box.decrypt(internal_message, nonce)
        return decrypted.decode('utf-8')



from abc import ABCMeta, abstractmethod


class BaseProvider(metaclass=ABCMeta):

    @abstractmethod
    async def restore_connection(self) -> None:
        raise NotImplementedError


    @abstractmethod
    def close_connection(self) -> None:
        raise NotImplementedError


    @abstractmethod
    async def disconnect(self) -> None:
        raise NotImplementedError


    @abstractmethod
    async def send_request(self, request) -> None:
        raise NotImplementedError


    @abstractmethod
    def listen(self, eventsCallback) -> None:
        raise NotImplementedError



import hashlib

from base64 import b64decode
from enum import IntEnum
from nacl.signing import VerifyKey
from nacl.encoding import HexEncoder

from pytonconnect.exceptions import ManifestContentError, ManifestNotFoundError, TonConnectError, UnknownError, BadRequestError, UnknownAppError, UserRejectsError
from pytonconnect.logger import _LOGGER


class CONNECT_EVENT_ERROR_CODES(IntEnum):
    UNKNOWN_ERROR = 0
    BAD_REQUEST_ERROR = 1
    MANIFEST_NOT_FOUND_ERROR = 2
    MANIFEST_CONTENT_ERROR = 3
    UNKNOWN_APP_ERROR = 100
    USER_REJECTS_ERROR = 300
    METHOD_NOT_SUPPORTED = 400


CONNECT_EVENT_ERRORS = {
    CONNECT_EVENT_ERROR_CODES.UNKNOWN_ERROR: UnknownError,
    CONNECT_EVENT_ERROR_CODES.BAD_REQUEST_ERROR: BadRequestError,
    CONNECT_EVENT_ERROR_CODES.UNKNOWN_APP_ERROR: UnknownAppError,
    CONNECT_EVENT_ERROR_CODES.USER_REJECTS_ERROR: UserRejectsError,
    CONNECT_EVENT_ERROR_CODES.USER_REJECTS_ERROR: ManifestNotFoundError,
    CONNECT_EVENT_ERROR_CODES.METHOD_NOT_SUPPORTED: ManifestContentError,
}


class CHAIN(IntEnum):
    MAINNET = '-239'
    TESTNET = '-3'


class DeviceInfo():

    platform: str # 'iphone' | 'ipad' | 'android' | 'windows' | 'mac' | 'linux' | 'browser'
    app_name: str # e.g. "Tonkeeper"
    app_version: str # e.g. "2.3.367"
    max_protocol_version: int
    features: list[dict]

    def from_dict(device: dict):
        device_info = DeviceInfo()
        device_info.platform = device['platform']
        device_info.app_name = device['appName']
        device_info.app_version = device['appVersion']
        device_info.max_protocol_version = device['maxProtocolVersion']
        device_info.features = device['features']
        return device_info


class Account():

    # User's address in "hex" format: "<wc>:<hex>"
    address: str

    # User's selected chain
    chain: CHAIN

    # Base64 (not url safe) encoded wallet contract state_init.
    # Can be used to get user's public key from the state_init if the wallet contract doesn't support corresponding method
    wallet_state_init: str

    # Hex string without 0x prefix
    public_key: str

    def __repr__(self):
        return f'<Account "{self.address}">'

    def from_dict(ton_addr: dict):
        if 'address' not in ton_addr:
            raise TonConnectError('address not contains in ton_addr')

        account = Account()
        account.address = ton_addr['address']
        account.chain = ton_addr['network']
        account.wallet_state_init = ton_addr['walletStateInit']
        account.public_key = ton_addr.get('publicKey', None)
        return account


class TonProof():

    timestamp: int
    domain_len: int
    domain_val: str
    payload: str
    signature: bytes

    def from_dict(reply: dict):
        proof = reply.get('proof', None)
        if proof is None:
            raise TonConnectError('proof not contains in ton_proof')

        ton_proof = TonProof()
        ton_proof.timestamp = proof['timestamp']
        ton_proof.domain_len = proof['domain']['lengthBytes']
        ton_proof.domain_val = proof['domain']['value']
        ton_proof.payload = proof['payload']
        ton_proof.signature = b64decode(proof['signature'])
        return ton_proof


class WalletInfo():

    # Information about user's wallet's device
    device: DeviceInfo

    # Provider type
    provider: str

    # Selected account
    account: Account

    # Response for ton_proof item request
    ton_proof: TonProof


    def __repr__(self):
        return f'<WalletInfo {self.account}>'


    def __init__(self):
        self.device = None
        self.provider = 'http' # only http supported
        self.account = None
        self.ton_proof = None


    def check_proof(self, src_payload: str=None) -> bool:
        if self.ton_proof is None:
            return False

        wc, whash = self.account.address.split(':', maxsplit=2)

        message = bytearray()
        message.extend('ton-proof-item-v2/'.encode())
        message.extend(int(wc, 10).to_bytes(4, 'little'))
        message.extend(bytes.fromhex(whash))
        message.extend(self.ton_proof.domain_len.to_bytes(4, 'little'))
        message.extend(self.ton_proof.domain_val.encode())
        message.extend(self.ton_proof.timestamp.to_bytes(8, 'little'))
        if src_payload is not None:
            message.extend(src_payload.encode())
        else:
            message.extend(self.ton_proof.payload.encode())

        signature_message = bytearray()
        signature_message.extend(bytes.fromhex('ffff'))
        signature_message.extend('ton-connect'.encode())
        signature_message.extend(hashlib.sha256(message).digest())

        try:
            verify_key = VerifyKey(self.account.public_key, HexEncoder)
            verify_key.verify(hashlib.sha256(signature_message).digest(), self.ton_proof.signature)
            _LOGGER.debug('PROOF IS OK')
            return True

        except Exception as e:
            _LOGGER.exception(f'PROOF ERROR')

        return False


class ConnectEventParser():

    def parse_response(payload: dict) -> WalletInfo:
        if 'items' not in payload:
            raise TonConnectError('items not contains in payload')

        wallet = WalletInfo()

        for item in payload['items']:
            if 'name' in item:
                if item['name'] == 'ton_addr':
                    wallet.account = Account.from_dict(item)
                elif item['name'] == 'ton_proof':
                    wallet.ton_proof = TonProof.from_dict(item)

        if wallet.account is None:
            raise TonConnectError('ton_addr not contains in items')

        wallet.device = DeviceInfo.from_dict(payload['device'])

        return wallet


    def parse_error(payload: dict) -> TonConnectError:
        error_constructor: TonConnectError = UnknownError

        code = payload.get('error', {}).get('code', None)
        if code is not None and code in CONNECT_EVENT_ERRORS:
            error_constructor = CONNECT_EVENT_ERRORS[code]
        
        message = payload.get('error', {}).get('message', None)
        return error_constructor(message)



import asyncio
import json
import requests

from aiohttp.client_exceptions import ClientConnectionError
from aiohttp_sse_client import client as sse_client

from pytonconnect.exceptions import TonConnectError
from pytonconnect.logger import _LOGGER
from pytonconnect.storage import IStorage


class BridgeGateway:

    SSE_PATH = 'events'
    POST_PATH = 'message'
    DEFAULT_TTL = 300

    _handle_listen: asyncio.Task
    _event_source: sse_client.EventSource
    _is_closed: bool

    _storage: IStorage
    _bridge_url: str
    _session_id: str
    _listener: any
    _errors_listener: any


    def __init__(self, storage: IStorage, bridge_url: str, session_id: str, listener, errors_listener):

        self._handle_listen = None
        self._event_source = None
        self._is_closed = False

        self._storage = storage
        self._bridge_url = bridge_url
        self._session_id = session_id
        self._listener = listener
        self._errors_listener = errors_listener
    

    async def listen_event_source(self, resolve: asyncio.Future):
        try:
            async with self._event_source:
                resolve.set_result(True)
                async for event in self._event_source:
                    await self._messages_handler(event)

        except asyncio.exceptions.TimeoutError:
            _LOGGER.exception(f'Bridge error -> TimeoutError')
        except asyncio.exceptions.CancelledError:
            pass
        except ClientConnectionError:
            _LOGGER.exception(f'Bridge error -> ClientConnectionError')
        except Exception:
            _LOGGER.exception(f'Bridge error -> Unknown')
        
        if not resolve.done():
            resolve.set_result(False)


    async def register_session(self) -> bool:
        if self._is_closed:
            return False

        bridge_base = self._bridge_url.rstrip('/')
        bridge_url = f'{bridge_base}/{self.SSE_PATH}?client_id={self._session_id}'
        
        last_event_id = await self._storage.get_item(IStorage.KEY_LAST_EVENT_ID)
        if last_event_id:
            bridge_url += f'&last_event_id={last_event_id}'
        _LOGGER.debug(f'Bridge url -> {bridge_url}')

        if self._handle_listen is not None:
            self._handle_listen.cancel()

        loop = asyncio.get_running_loop()
        resolve = loop.create_future()

        self._event_source = sse_client.EventSource(bridge_url, timeout=-1, on_error=self._errors_handler)
        self._handle_listen = asyncio.create_task(self.listen_event_source(resolve))

        return await resolve


    async def send(self, request: str, receiver_public_key: str, topic: str, ttl: int = None):
        bridge_base = self._bridge_url.rstrip('/')
        bridge_url = f'{bridge_base}/{self.POST_PATH}?client_id={self._session_id}'
        bridge_url += f'&to={receiver_public_key}'
        bridge_url += f'&ttl={ttl if ttl else self.DEFAULT_TTL}'
        bridge_url += f'&topic={topic}'
        requests.post(bridge_url, request)


    def pause(self):
        if self._handle_listen is not None:
            self._handle_listen.cancel()
            self._handle_listen = None


    async def unpause(self):
        await self.register_session()


    def close(self):
        self._is_closed = True
        self.pause()


    async def _messages_handler(self, event: sse_client.MessageEvent):
        await self._storage.set_item(IStorage.KEY_LAST_EVENT_ID, event.last_event_id)

        if not self._is_closed:
            try:
                bridge_incoming_message = json.loads(event.data)
            except:
                raise TonConnectError(f'Bridge message parse failed, message {event.data}')
            else:
                await self._listener(bridge_incoming_message)


    def _errors_handler(self):
        if not self._is_closed:
            if self._event_source.ready_state == sse_client.READY_STATE_CLOSED:
                _LOGGER.error('Bridge error -> READY_STATE_CLOSED')
                # TODO: reconnect
                return

            elif self._event_source.ready_state == sse_client.READY_STATE_CONNECTING:
                _LOGGER.error('Bridge error -> READY_STATE_CONNECTING')
                return
            
            if not self._errors_listener:
                self._errors_listener()



from pytonconnect.exceptions import ManifestContentError, ManifestNotFoundError, WalletAlreadyConnectedError, WalletNotConnectedError, WalletNotSupportFeatureError
from pytonconnect.logger import _LOGGER
from pytonconnect.parsers import SendTransactionParser, ConnectEventParser, WalletInfo
from pytonconnect.provider import BridgeProvider
from pytonconnect.storage import IStorage, DefaultStorage

from ._wallets_list_manager import WalletsListManager


class TonConnect:

    _wallets_list = WalletsListManager()

    _provider: BridgeProvider
    _manifest_url: str
    _storage: IStorage

    _wallet: WalletInfo

    _status_change_subscriptions: list
    _status_change_error_subscriptions: list

    @property
    def connected(self):
        """Shows if the wallet is connected right now."""
        return self._wallet is not None

    @property
    def account(self):
        """Current connected account or None if no account is connected."""
        return self._wallet.account if self.connected else None

    @property
    def wallet(self):
        """Current connected wallet or None if no account is connected."""
        return self._wallet


    def __init__(self, manifest_url: str, storage: IStorage=DefaultStorage(), wallets_list_source: str = None, wallets_list_cache_ttl: int = None):
        self._wallets_list = WalletsListManager(wallets_list_source=wallets_list_source, cache_ttl=wallets_list_cache_ttl)

        self._provider = None
        self._manifest_url = manifest_url
        self._storage = storage

        self._wallet = None

        self._status_change_subscriptions = []
        self._status_change_error_subscriptions = []

    def get_wallets(self):
        """Return available wallets list."""
        return self._wallets_list.get_wallets()


    def on_status_change(self, callback, errors_handler = None):
        """Allows to subscribe to connection status changes and handle connection errors.
        
        :param callback: will be called after connections status changes with actual wallet or None
        :param errors_handler: will be called with some instance of TonConnectError when connect error is received
        :return: unsubscribe callback
        """
        self._status_change_subscriptions.append(callback)
        if errors_handler is not None:
            self._status_change_error_subscriptions.append(errors_handler)

        def unsubscribe():
            if callback in self._status_change_subscriptions:
                self._status_change_subscriptions.remove(callback)
            if errors_handler is not None and errors_handler in self._status_change_error_subscriptions:
                self._status_change_error_subscriptions.remove(errors_handler)

        return unsubscribe


    async def connect(self, wallet, request = None):
        """Generates universal link for an external wallet and subscribes to the wallet's bridge, or sends connect request to the injected wallet.

        :param wallet: wallet's bridge url and universal link for an external wallet.
        :param request: additional request to pass to the wallet while connect (currently only ton_proof is available).
        :return: universal link if external wallet was passed.
        """
        if self.connected:
            raise WalletAlreadyConnectedError()

        if self._provider:
            self._provider.close_connection()
        
        self._provider = self._create_provider(wallet)

        return await self._provider.connect(self._create_connect_request(request))


    async def restore_connection(self):
        """Try to restore existing session and reconnect to the corresponding wallet. Call it immediately when your app is loaded.

        :return: True if connection is restored
        """
        try:
            self._provider = BridgeProvider(self._storage)
        except:
            await self._storage.remove_item(IStorage.KEY_CONNECTION)
            self._provider = None
        
        if not self._provider:
            return False

        self._provider.listen(self._wallet_events_listener)
        return await self._provider.restore_connection()


    async def send_transaction(self, transaction: dict):
        """Asks connected wallet to sign and send the transaction.
        
        :param transaction: transaction to send.
        :return: signed transaction boc that allows you to find the transaction in the blockchain.
        If user rejects transaction, method will throw the corresponding error.
        """
        if not self.connected:
            raise WalletNotConnectedError()

        options = {'required_messages_number': len(transaction.get('messages', []))}
        self._check_send_transaction_support(self._wallet.device.features, options)

        request = {
            'valid_until': transaction.get('valid_until', None),
            'from': transaction.get('from', self._wallet.account.address),
            'network': transaction.get('network', self._wallet.account.chain),
            'messages': transaction.get('messages', [])
        }

        response = await self._provider.send_request(SendTransactionParser.convert_to_rpc_request(request))

        if SendTransactionParser.is_error(response):
            return SendTransactionParser.parse_and_throw_error(response)

        return SendTransactionParser.convert_from_rpc_response(response)


    async def disconnect(self):
        """Disconnect from wallet and drop current session."""
        if not self.connected:
            raise WalletNotConnectedError()

        await self._provider.disconnect()
        self._on_wallet_disconnected()


    def pause_connection(self):
        """Pause bridge HTTP connection. Might be helpful, if you use SDK on backend and want to save server resources."""
        self._provider.pause()

    
    async def unpause_connection(self):
        """Unpause bridge HTTP connection if it is paused."""
        await self._provider.unpause()
    

    def _check_send_transaction_support(self, features, options):
        supports_deprecated_send_transaction_feature = 'SendTransaction' in features
        send_transaction_feature = None
        for feature in features:
            if isinstance(feature, dict) and feature.get('name', None) == 'SendTransaction':
                send_transaction_feature = feature
                break

        if not supports_deprecated_send_transaction_feature and not send_transaction_feature:
            raise WalletNotSupportFeatureError("Wallet doesn't support SendTransaction feature.")

        if send_transaction_feature:
            max_messages = send_transaction_feature.get('maxMessages', None)
            required_messages = options.get('required_messages_number')
            if max_messages and max_messages < required_messages:
                raise WalletNotSupportFeatureError(f'Wallet is not able to handle such SendTransaction request. Max support messages number is {max_messages}, but {required_messages} is required.')
        else:
            _LOGGER.warning("Connected wallet didn't provide information about max allowed messages in the SendTransaction request. Request may be rejected by the wallet.")


    def _create_provider(self, wallet: dict) -> BridgeProvider:
        provider = BridgeProvider(self._storage, wallet)
        provider.listen(self._wallet_events_listener)
        return provider


    def _wallet_events_listener(self, data):
        if data['event'] == 'connect':
            self._on_wallet_connected(data['payload'])

        elif data['event'] == 'connect_error':
            self._on_wallet_connect_error(data['payload'])

        elif data['event'] == 'disconnect':
            self._on_wallet_disconnected()


    def _on_wallet_connected(self, payload):
        self._wallet = ConnectEventParser.parse_response(payload)
        for listener in self._status_change_subscriptions:
            listener(self._wallet)


    def _on_wallet_connect_error(self, payload):
        _LOGGER.debug('connect error %s', payload)
        error = ConnectEventParser.parse_error(payload)
        for listener in self._status_change_error_subscriptions:
            listener(error)
        
        if isinstance(error, ManifestNotFoundError) or isinstance(error, ManifestContentError):
            _LOGGER.exception(error)
            raise error


    def _on_wallet_disconnected(self):
        self._wallet = None
        for listener in self._status_change_subscriptions:
            listener(None)


    def _create_connect_request(self, request):
        items = [
            {
                'name': 'ton_addr'
            }
        ]

        if isinstance(request, dict) and 'ton_proof' in request:
            items.append({
                'name': 'ton_proof',
                'payload': request['ton_proof']
            })

        return {
            'manifestUrl': self._manifest_url,
            'items': items
        }



from ._interface import IStorage


class DefaultStorage(IStorage):

    _cache: dict

    def __init__(self):
        self._cache = {}


    async def set_item(self, key: str, value: str):
        self._cache[key] = value


    async def get_item(self, key: str, default_value: str = None):
        if key not in self._cache:
            return default_value
        return self._cache[key]


    async def remove_item(self, key: str):
        if key in self._cache:
            del self._cache[key]



import json

from enum import IntEnum

from pytonconnect.exceptions import TonConnectError, UnknownError, BadRequestError, UnknownAppError, UserRejectsError
from ._rpc_parser import RpcParser


class SEND_TRANSACTION_ERROR_CODES(IntEnum):
    UNKNOWN_ERROR = 0
    BAD_REQUEST_ERROR = 1
    UNKNOWN_APP_ERROR = 100
    USER_REJECTS_ERROR = 300
    METHOD_NOT_SUPPORTED = 400


SEND_TRANSACTION_ERRORS = {
    SEND_TRANSACTION_ERROR_CODES.UNKNOWN_ERROR: UnknownError,
    SEND_TRANSACTION_ERROR_CODES.BAD_REQUEST_ERROR: BadRequestError,
    SEND_TRANSACTION_ERROR_CODES.UNKNOWN_APP_ERROR: UnknownAppError,
    SEND_TRANSACTION_ERROR_CODES.USER_REJECTS_ERROR: UserRejectsError,
}


class SendTransactionParser(RpcParser):
    
    def convert_to_rpc_request(request: dict) -> dict:
        return {
            'method': 'sendTransaction',
            'params': [json.dumps(request)]
        }


    def convert_from_rpc_response(rpc_response: dict) -> dict:
        return {
            'boc': rpc_response['result']
        }


    def parse_and_throw_error(response: dict) -> None:
        error_constructor: TonConnectError = UnknownError

        code = response.get('error', {}).get('code', None)
        if code is not None and code in SEND_TRANSACTION_ERRORS:
            error_constructor = SEND_TRANSACTION_ERRORS[code]
        
        message = response.get('error', {}).get('message', None)
        raise error_constructor(message)
